import csv
import logging
from knnslim.graph import graph_utils

__all__ = ['Dependency', 'AddDependency', 'CatDependency', 'InputDependency', 'OutputDependency']

CONV_TYPES = ['Conv1d', 'Conv2d', 'Conv3d', 
              'ConvTranspose1d', 'ConvTranspose2d', 'ConvTranspose3d',
              'aten::_convolution', 
              'Conv2dQuant', 'ConvTranspose2dQuant']
FC_TYPES = ['Linear', 'aten::addmm', 'aten::matmul', "aten::linear",
            'LinearQuant']
ADD_TYPES = ['aten::add', 'aten::add_', 'aten::div', 'aten::bmm', 'aten::mul', 
             'AvgQuant', 'MulQuant']
CAT_TYPES = ['aten::cat']
CHUNK_TYPES = ['aten::chunk']
RESHAPE_OPS = ['aten::view', 'aten::reshape', 'aten::flatten', 'aten::mean', 'aten::squeeze_', 'aten::transpose']
OTHER_OPS = ['aten::Int', 'aten::size']
logger = logging.getLogger('Model Dependency')


class Dependency:
    def __init__(self, model=None, input=None, traced_model=None):
        """
        Build the graph for the model.
        """
        if traced_model is not None:
            print('-> Analyze [%s] via "traced_model"' % model.name)
        elif model is not None and input is not None:
            print('-> Analyze [%s] via "model" & "input"' % model.name)
        else:
            raise Exception('Analyze requires one of these: traced_model, model & input')

        self.module_graph = graph_utils.TorchModuleGraph(model, input, traced_model)
        self.dp_dict = dict()
        self.full_dp_dict = dict()
        self.build()

    def build(self):
        pass

    def export(self, filepath):
        pass

    def get_traced_model(self):
        return self.module_graph.get_traced_model()

    def get_name_to_node(self):
        return self.module_graph.name_to_node

    def get_leaf_module_names(self):
        return self.module_graph.leaf_module_names

    def get_name_to_module(self):
        return self.module_graph.name_to_module

    def get_name_to_type(self, name):
        return self.module_graph.name_to_node[name].op_type

    def _get_prev_key_layers(self, node, omit_types=()):
        """
        Find the nearest previous conv/linear layers for the target node.

        @Parameters
        ------
        node : (torch._C.Node)
            target node.
        omit_types : (tuple of names)
            op_types to be omitted in searching

        @Returns
        ------
        prev_layers: (list)
            nearest previous conv/linear layers for the target node.
        """
        key_types = CONV_TYPES + FC_TYPES + ADD_TYPES + CAT_TYPES + CHUNK_TYPES
        key_types = [x for x in key_types if x not in omit_types]
        #print(key_types)
        
        skip_types = OTHER_OPS
        prev_layers = []
        prevs = self.module_graph.find_predecessors(node.unique_name)
        prevs = [self.module_graph.name_to_node[name] for name in prevs]
        queue = [node for node in prevs]
        while queue:
            cur_node = queue.pop(0)
            if cur_node.op_type in key_types:
                # find the first met conv/linear
                prev_layers.append(cur_node.name)
                continue
            prevs = self.module_graph.find_predecessors(cur_node.unique_name)
            prevs = [self.module_graph.name_to_node[name] for name in prevs]
            prevs = [x for x in prevs if x not in queue and x.op_type not in skip_types]
            queue = prevs + queue
        prev_layers = sorted(set(prev_layers), key=prev_layers.index)
        return prev_layers

    def _get_next_key_layers(self, node, omit_types=()):
        """
        Find the nearest subsequent conv/linear layers for the target node.

        @Parameters
        ------
        node : (torch._C.Node)
            target node.
        omit_types : (tuple of names)
            op_types to be omitted in searching

        @Returns
        ------
        next_layers: (list)
            nearest subsequent conv/linear layers for the target node.
        """
        key_types = CONV_TYPES + FC_TYPES
        key_types = [x for x in key_types if x not in omit_types]
        skip_types = OTHER_OPS + ADD_TYPES + CAT_TYPES + CHUNK_TYPES
        next_layers = []
        queue = [node]
        while queue:
            cur_node = queue.pop(0)
            if cur_node.op_type in key_types:
                # find the first met conv/linear
                next_layers.append(cur_node.name)
                continue
            nexts = self.module_graph.find_successors(cur_node.unique_name)
            nexts = [self.module_graph.name_to_node[name] for name in nexts]
            nexts = [x for x in nexts if x not in queue and x.op_type not in skip_types]
            queue = nexts + queue
        next_layers = sorted(set(next_layers), key=next_layers.index)
        return next_layers

    def _get_prev_layers(self, node):
        """
        Find the nearest previous layers for the target node.

        @Parameters
        ------
        node : (torch._C.Node)
            target node.

        @Returns
        ------
        prev_layers: (list)
            nearest previous layers for the target node.
        """
        key_types = CONV_TYPES + FC_TYPES
        end_types = ADD_TYPES + CAT_TYPES + CHUNK_TYPES
        skip_types = OTHER_OPS
        prev_layers = []
        prevs = self.module_graph.find_predecessors(node.unique_name)
        prevs = [self.module_graph.name_to_node[name] for name in prevs]
        queue = [node for node in prevs]
        while queue:
            cur_node = queue.pop(0)
            if cur_node.op_type in end_types:
                continue
            prev_layers.append(cur_node.name)
            if cur_node.op_type in key_types:
                # find the first met conv/linear
                continue
            prevs = self.module_graph.find_predecessors(cur_node.unique_name)
            prevs = [self.module_graph.name_to_node[name] for name in prevs]
            prevs = [x for x in prevs if x not in queue and x.op_type not in skip_types]
            queue = prevs + queue
        prev_layers = sorted(set(prev_layers), key=prev_layers.index)
        return prev_layers

    def _get_next_layers(self, node):
        """
        Find the nearest subsequent layers for the target node.

        @Parameters
        ------
        node : (torch._C.Node)
            target node.

        @Returns
        ------
        next_layers: (list)
            nearest subsequent layers for the target node.
        """
        key_types = CONV_TYPES + FC_TYPES
        end_types = ADD_TYPES + CAT_TYPES + CHUNK_TYPES
        skip_types = OTHER_OPS
        next_layers = []
        nexts = self.module_graph.find_successors(node.unique_name)
        nexts = [self.module_graph.name_to_node[name] for name in nexts]
        queue = [node for node in nexts]
        while queue:
            cur_node = queue.pop(0)
            if cur_node.op_type in end_types:
                continue
            next_layers.append(cur_node.name)
            if cur_node.op_type in key_types:
                # find the first met conv/linear
                continue
            nexts = self.module_graph.find_successors(cur_node.unique_name)
            nexts = [self.module_graph.name_to_node[name] for name in nexts]
            nexts = [x for x in nexts if x not in queue and x.op_type not in skip_types]
            queue = nexts + queue
        next_layers = sorted(set(next_layers), key=next_layers.index)
        return next_layers


class AddDependency(Dependency):
    def __init__(self, model=None, input=None, traced_model=None):
        """
        Analyze the channel dependencies around "Add" in a torch model

        @Parameters
        ------
        model : (torch.nn.Module)
            The model to be analyzed.
        input : (torch.Tensor)
            The example input data to trace the network architecture.
        traced_model : (torch._C.Graph)
            if we already has the traced graph of the target model, we don't
            need to trace the model again.
        """
        super(AddDependency, self).__init__(model, input, traced_model)
        self.dependency_type = 'add'

    def build(self):
        """
        Build the add dependencies. Adjacent Add are merged to share the same dependencies.
        """
        # unpack the tuple/list manually before analyze the channel dependency
        self.module_graph.unpack_manually()
        # print('nodes_op', self.module_graph.nodes_py.nodes_op)
        for node in self.module_graph.nodes_py.nodes_op:
            # print(node.name,node.op_type)
            prev_layers = []
            next_layers = []
            # find the node that is Add or Cat types
            if node.op_type in ADD_TYPES:
                # use DFS to find prev/next layers
                prev_layers = self._get_prev_key_layers(node, omit_types=ADD_TYPES)
                next_layers = self._get_next_key_layers(node, omit_types=ADD_TYPES)
            elif node.op_type in CAT_TYPES:
                # To determine if this cat operation will introduce channel
                # dependency, we need the specific input parameters of the cat
                # operation. To get the input parameters of the cat operation, we
                # need to traverse all the cpp_nodes included by this NodePyGroup,
                # because, graph_utils.TorchModuleGraph merges the important nodes and the adjacent
                # unimportant nodes (nodes started with prim::attr, for example) into a
                # NodepyGroup.
                cat_dim = None
                for cnode in node.node_cpps:
                    if cnode.kind() in CAT_TYPES:
                        cat_dim = list(cnode.inputs())[1].toIValue()
                        break
                if cat_dim != 1:
                    prev_layers = self._get_prev_key_layers(node, omit_types=CAT_TYPES)
            # merge the dependencies
            tmp_set0 = []
            tmp_set1 = []
            tmp_set2 = []
            for _name in prev_layers:
                if _name in self.full_dp_dict:
                    tmp_set0 += self.full_dp_dict[_name][0]
                    tmp_set1 += self.full_dp_dict[_name][1]
                    tmp_set2 += self.full_dp_dict[_name][2]
            tmp_set0 += prev_layers  # merge previous layers
            tmp_set1 += [node.name, ]  # merge add names
            tmp_set2 += next_layers  # merge subsequent layers
            full_prev_layers = sorted(set(tmp_set0), key=tmp_set0.index)  # remove redundant names
            full_node_names = sorted(set(tmp_set1), key=tmp_set1.index)
            full_next_layers = sorted(set(tmp_set2), key=tmp_set2.index)
            # save the dependencies
            for _name in full_prev_layers:
                if len(full_prev_layers) <= 1: continue
                self.full_dp_dict[_name] = [full_prev_layers, full_node_names, full_next_layers]
            for _name in full_node_names:
                if full_prev_layers:
                    self.dp_dict[_name] = [full_prev_layers, full_next_layers]

    def dependency_list(self, with_type=False):
        """
        Get the list of the clean dependency set.

        @Parameters
        ------
        with_type : (bool)
            whether to give out the op_type information

        @Returns
        ------
        d_list : (list)
            [(prev_layers), (add_names), (next_layers)}, e.g.,
            [('conv1', 'conv2'), ('add_1'), ('conv3', 'conv4')]
        """
        d_list = []
        for info in self.full_dp_dict.values():
            prev_names, add_names, next_names = tuple(info[0]), tuple(info[1]), tuple(info[2])
            d_list.append(tuple([prev_names, add_names, next_names]))
        d_list = sorted(set(d_list), key=d_list.index)
        return d_list

    def dependency_dict(self, with_type=False):
        """
        Get the dict of the individual dependency dict.

        @Parameters
        ------
        with_type : (bool)
            whether to give out the op_type information

        @Returns
        ------
        d_dict : (dict)
            {key=name, value=[(prev_layers), (next_layers)]}, e.g.,
            {'add_name' : [('conv1', 'conv2'), ('conv3', 'conv4')]}
        """
        d_dict = {}
        for add_name, (prev_names, next_names) in self.dp_dict.items():
            if with_type:
                add_name = '{}({})'.format(add_name, self.get_name_to_type(add_name))
                prev_names = ['{}({})'.format(x, self.get_name_to_type(x)) for x in prev_names]
                next_names = ['{}({})'.format(x, self.get_name_to_type(x)) for x in next_names]
            d_dict[add_name] = [tuple(prev_names), tuple(next_names)]
        return d_dict

    def export(self, filepath, with_type=False):
        """
        Export the channel dependencies as a csv file.
        The layers at the same line have output channel dependencies with each other,
        which means the output channel (filters) numbers of three layers should be same
        with each other, otherwise the model may has shape conflict.

        Add Dependency Set
        [#id,add_name,(prev_layers),(next_layers)], e.g.,
        [
            [1,'add_1',('conv1','conv2'),('conv3')],
            [2,'add_2',('conv4','conv5'),('conv6')],
            [3,'add_3',('conv7','conv8'),('conv9')],
        ]
        """
        header = ['#id', 'Layer', '(Before Add)', '[After Add]']
        with open(filepath, 'w') as csvf:
            csv_w = csv.writer(csvf, delimiter=',')
            csv_w.writerow(header)
            for id, (name, info) in enumerate(self.dependency_dict(with_type).items()):
                prev_add, next_add = info
                row = [id, name, tuple(prev_add), tuple(next_add)]
                csv_w.writerow(row)


# def reshape_break_channel_dependency(op_node):
#     """
#     The reshape operations such as (reshape, view, flatten) may break
#     the channel dependency. We need to check the input parameters of
#     these reshape operations to check if this reshape node will break
#     the channel dependency. However, it's complicated to analyze the the input
#     parameters for each reshape function and infer if it will break the channel
#     dependency. So currently, we just check if the input channel and the output
#     channel is the same, if so, then we can say the original reshape function
#     doesn't want to change the number of the channels, which means the channel
#     dependency is not broken. In contrast, the original reshape operation wants
#     to change the number of channels, so it breaks the channel dependency.
#
#     @Parameters
#     ------
#     op_node: NodePyOP
#         A Op node of the graph.
#     Returns
#     ------
#     bool
#         If this operation will break the channel dependency.
#     """
#     in_shape = op_node.auxiliary['in_shape']
#     out_shape = op_node.auxiliary['out_shape']
#     in_channel = in_shape[1]
#     out_channel = out_shape[1]
#     return in_channel != out_channel
#
#
# class InputChannelDependency(Dependency):
#     """
#     Some pruners may prune the input channel of the convolutional
#     layers. While pruning the input channel of the convolutional layers,
#     the layers that share the same input tensor should prune the same
#     channels, and we say these layers that share the same input tensor/channel
#     has the input channel dependency. If we only prune the input channel of one
#     layer in the dependency set, there will be a shape conflict for the other
#     layers in the same dependency set, which may trigger a runtime error.
#     Here we judge whether the application will truncate the dependency by analyzing
#     whether the number of channels before and after the operation has changed.
#     If not, the input channel dependency will be passed to the following nodes.
#     """
#
#     def __init__(self, model, input=None, traced_model=None):
#         """
#         This model analyze the input channel dependencies between the conv
#         layers in a model.
#
#         @Parameters
#         ------
#         model : torch.nn.Module
#             The model to be analyzed.
#         input : torch.Tensor
#             The example input data to trace the network architecture.
#         traced_model : torch._C.Graph
#             if we already has the traced graph of the target model, we don't
#             need to trace the model again.
#         """
#         super(InputChannelDependency, self).__init__(model, input, traced_model)
#
#     def _get_following_convs(self, tensor):
#         queue = []
#         key_layers = []
#         queue.extend(self.graph.input_to_node[tensor])
#         op_types = CONV_TYPES + FC_TYPES
#         while queue:
#             cur_node = queue.pop(0)
#             if cur_node.op_type in op_types:
#                 # find the first met conv
#                 key_layers.append(cur_node.name)
#                 continue
#             elif cur_node.op_type in RESHAPE_OPS:
#                 # check if the reshape operation will break the channel dependency
#                 if reshape_break_channel_dependency(cur_node):
#                     # reshape operations also breaks the dependency relationship
#                     continue
#             successors = self.graph.find_successors(cur_node.unique_name)
#             successors = [self.graph.name_to_node[name] for name in successors]
#             for layer in successors:
#                 queue.append(layer)
#         return key_layers
#
#     def build(self):
#         """
#         Build the input channel dependencies.
#         The `InputChannelDependency` indicates the layers that have
#         dependencies when pruning the input channel of the conv layers.
#         In contrast, `ChannelDependency` indicates the dependent layers
#         when pruning the output channels of conv layers (for example, L1FilterPruner).
#         """
#         # unpack the tuple or list manually
#         self.graph.unpack_manually()
#         for tensor in self.graph.input_to_node:
#             # start from this tensor, find all the conv layers that
#             # take this tensor as input. Similar to the `ChannelDependency`
#             # the conv layer will truncate the dependencies
#             layers = self._get_following_convs(tensor)
#             dependency_set = set(layers)
#             for layer in layers:
#                 if layer in self.dependency:
#                     dependency_set.update(self.dependency[layer])
#             for layer in dependency_set:
#                 self.dependency[layer] = dependency_set


class CatDependency(Dependency):
    def __init__(self, model=None, input=None, traced_model=None):
        """
        Analyze the channel dependencies around "Cat" in a torch model

        @Parameters
        ------
        model : (torch.nn.Module)
            The model to be analyzed.
        input : (torch.Tensor)
            The example input data to trace the network architecture.
        traced_model : (torch._C.Graph)
            if we already has the traced graph of the target model, we don't
            need to trace the model again.
        """
        super(CatDependency, self).__init__(model, input, traced_model)

    def build(self):
        """
        Build the concat dependencies. Several adjacent Cat have not dependencies.
        """
        # unpack the tuple or list manually
        self.module_graph.unpack_manually()
        for node in self.module_graph.nodes_py.nodes_op:
            prev_layers = []
            next_layers = []
            if node.op_type in CAT_TYPES and node.auxiliary['cat_dim'] == 1:
                prev_layers = self._get_prev_key_layers(node)
                next_layers = self._get_next_key_layers(node)
                # do not merge dependencies for concat
                full_prev_layers = prev_layers
                full_next_layers = next_layers
                # save the dependencies
                # for _name in full_prev_layers:
                #     self.full_dp_dict[_name] = [full_prev_layers, [node.name], full_next_layers]
                if full_prev_layers:
                    self.dp_dict[node.name] = [full_prev_layers, full_next_layers]

    def dependency_list(self, with_type=False):
        """
        Get the list of the clean dependency set.

        @Parameters
        ------
        with_type : (bool)
            whether to give out the op_type information

        @Returns
        ------
        d_list : (list)
            [(prev_layers), (cat_names), (next_layers)}, e.g.,
            [('conv1', 'conv2'), ('cat_1'), ('conv3', 'conv4')]
        """
        d_list = []
        for name, (prev_names, next_names) in self.dp_dict.items():
            if with_type:
                prev_names = ['{}({})'.format(x, self.get_name_to_type(x)) for x in prev_names]
                next_names = ['{}({})'.format(x, self.get_name_to_type(x)) for x in next_names]
            d_list.append(tuple([tuple(prev_names), tuple([name]), tuple(next_names)]))
        return d_list

    def dependency_dict(self, with_type=False):
        """
        Get the dict of the individual dependency dict.

        @Parameters
        ------
        with_type : (bool)
            whether to give out the op_type information

        @Returns
        ------
        d_dict : (dict)
            {key=name, value=[(prev_layers), (next_layers)]}, e.g.,
            {cat_name : [('conv1', 'conv2'), ('conv3', 'conv4')]}
        """
        d_dict = {}
        for name, (prev_names, next_names) in self.dp_dict.items():
            if with_type:
                prev_names = ['{}({})'.format(x, self.get_name_to_type(x)) for x in prev_names]
                next_names = ['{}({})'.format(x, self.get_name_to_type(x)) for x in next_names]
            d_dict[name] = [tuple(prev_names), tuple(next_names)]
        return d_dict

    def export(self, filepath, with_type=False):
        """
        Export the channel dependencies as a csv file.
        The layers at the same line have output channel dependencies.
        Each line contains layers whose output channels are concatenated together. Likewise,
        there are layers whose input channels are concatenated together.

        Cat Dependency Set
        [#id,cat_name,(prev_layers),(next_layers)], e.g.,
        [
            [1,'cat_1',('conv1','conv2'),('conv3')],
            [2,'cat_2',('conv4','conv5'),('conv6')],
            [3,'cat_3',('conv7','conv8'),('conv9')],
        ]
        """
        header = ['#id', 'Layer', '(Before Cat)', '[After Cat]']
        with open(filepath, 'w') as csvf:
            csv_w = csv.writer(csvf, delimiter=',')
            csv_w.writerow(header)
            for id, (name, info) in enumerate(self.dependency_dict(with_type).items()):
                prev_cat, next_cat = info
                row = [id, name, tuple(prev_cat), tuple(next_cat)]
                csv_w.writerow(row)


class InputDependency(Dependency):
    def __init__(self, model=None, input=None, traced_model=None):
        """
        Analyze the input dependencies given a node in a torch model

        @Parameters
        ------
        model : (torch.nn.Module)
            The model to be analyzed.
        input : (torch.Tensor)
            The example input data to trace the network architecture.
        traced_model : (torch._C.Graph)
            if we already has the traced graph of the target model, we don't
            need to trace the model again.
        """
        super(InputDependency, self).__init__(model, input, traced_model)

    def build(self):
        """
        Build the input dependencies.
        """
        # unpack the tuple/list manually before analyze the
        # channel dependency
        self.module_graph.unpack_manually()
        key_ops = CONV_TYPES + FC_TYPES + ADD_TYPES + CAT_TYPES + CHUNK_TYPES
        for node in self.module_graph.nodes_py.nodes_op:
            prev_layers = []
            if node.op_type in key_ops:
                prev_layers = self._get_prev_layers(node)
                # save the dependencies
                self.dp_dict[node.name] = prev_layers

    def dependency_list(self, with_type=False):
        d_dict = self.dependency_dict(with_type)
        d_list = [(name, info) for name, info in d_dict.items()]
        return d_list

    def dependency_dict(self, with_type=False):
        """
        Get the dict of the clean dependency set.

        @Parameters
        ------
        with_type : (bool)
            whether to give out the op_type information

        @Returns
        ------
        d_set : (dict)
            {key=name, value=(prev_layers)}, e.g.,
            {'conv2' : ('relu1', 'batchnorm1', 'conv1')}
        """
        d_dict = {}
        for name, info in self.dp_dict.items():
            if with_type:
                name = '{}({})'.format(name, self.get_name_to_type(name))
                info = ['{}({})'.format(x, self.get_name_to_type(x)) for x in info]
            d_dict[name] = tuple(info)
        return d_dict

    def export(self, filepath, with_type=False):
        """
        Export the channel dependencies as a csv file.

        Input Dependency Set,
        [#id,name,(prev_layers)], e.g.,
        [
            [1,'conv2',('relu1','batchnorm1','conv1')],
            [2,'conv3',('relu2','batchnorm2','conv2')],
        ]
        """
        header = ['#id', 'Layer', '[Prev layers]']
        with open(filepath, 'w') as csvf:
            csv_w = csv.writer(csvf, delimiter=',')
            csv_w.writerow(header)
            for id, (name, info) in enumerate(self.dependency_dict(with_type).items()):
                row = [id, name, info]
                csv_w.writerow(row)


class OutputDependency(Dependency):
    def __init__(self, model=None, input=None, traced_model=None):
        """
        Analyze the output dependencies given a node in a torch model

        @Parameters
        ------
        model : (torch.nn.Module)
            The model to be analyzed.
        input : (torch.Tensor)
            The example input data to trace the network architecture.
        traced_model : (torch._C.Graph)
            if we already has the traced graph of the target model, we don't
            need to trace the model again.
        """
        super(OutputDependency, self).__init__(model, input, traced_model)

    def build(self):
        """
        Build the output dependencies.
        """
        # unpack the tuple/list manually before analyze the
        # channel dependency
        self.module_graph.unpack_manually()
        key_ops = CONV_TYPES + FC_TYPES + ADD_TYPES + CAT_TYPES + CHUNK_TYPES
        for node in self.module_graph.nodes_py.nodes_op:
            next_layers = []
            if node.op_type in key_ops:
                next_layers = self._get_next_layers(node)
                # save the dependencies
                self.dp_dict[node.name] = next_layers

    def dependency_list(self, with_type=False):
        d_dict = self.dependency_dict(with_type)
        d_list = [(name, info) for name, info in d_dict.items()]
        return d_list

    def dependency_dict(self, with_type=False):
        """
        Get the dict of the clean dependency set.

        @Parameters
        ------
        with_type : (bool)
            whether to give out the op_type information

        @Returns
        ------
        d_set : (dict)
            {key=name, value=(next_layers)}, e.g.,
            {'conv1' : ('batchnorm1', 'relu1', 'conv2')}
        """
        d_dict = {}
        for name, info in self.dp_dict.items():
            if with_type:
                name = '{}({})'.format(name, self.get_name_to_type(name))
                info = ['{}({})'.format(x, self.get_name_to_type(x)) for x in info]
            d_dict[name] = tuple(info)
        return d_dict

    def export(self, filepath, with_type=False):
        """
        Export the channel dependencies as a csv file.

        Output Dependency Set,
        [#id,name,(next_layers)], e.g.,
        [
            [1,'conv1',('batchnorm1','relu1','conv2')],
            [2,'conv2',('batchnorm2','relu2','conv3')],
        ]
        """
        header = ['#id', 'Layer', '[Next layers]']
        with open(filepath, 'w') as csvf:
            csv_w = csv.writer(csvf, delimiter=',')
            csv_w.writerow(header)
            for id, (name, info) in enumerate(self.dependency_dict(with_type).items()):
                row = [id, name, info]
                csv_w.writerow(row)

# class GroupDependency(Dependency):
#     def __init__(self, model=None, input=None, traced_model=None):
#         """
#         This model analyze the group dependencies between the conv
#         layers in a model.
#
#         @Parameters
#         ------
#         model : torch.nn.Module
#             The model to be analyzed.
#         input : torch.Tensor
#             The example input data to trace the network architecture.
#         traced_model : torch._C.Graph
#             if we already has the traced graph of the target model, we don't
#             need to trace the model again.
#         """
#         super(GroupDependency, self).__init__(model, input, traced_model)
#
#     def _get_parent_convs(self, node):
#         """
#         Find the nearest father conv layers for the target node.
#
#         @Parameters
#         ------
#         node : torch._C.Node
#             target node.
#
#         @Returns
#         ------
#         parent_layers : list
#             nearest father conv layers for the target node. Due to the group
#             dependency only exists between the conv layers, so we only find
#             the parent conv layers.
#         """
#         parent_layers = []
#         # the input node is a Conv node
#         predecessors = self.graph.find_predecessors(node.unique_name)
#         predecessors = [self.graph.name_to_node[x] for x in predecessors]
#         queue = predecessors
#         while queue:
#             cur_node = queue.pop(0)
#             if cur_node.op_type in ['Conv2d', 'ConvTranspose2d', CONV_TYPE]:
#                 # find the first met conv
#                 parent_layers.append(cur_node.name)
#                 continue
#             parents = self.graph.find_predecessors(cur_node.unique_name)
#             parents = [self.graph.name_to_node[name] for name in parents]
#             for parent in parents:
#                 queue.append(parent)
#         return parent_layers
#
#     def _get_conv_groups(self, node_group):
#         """
#         Get the number of groups for a convolutional layer.
#
#         @Parameters
#         ------
#         node_group : NodePyGroup
#             target node.
#
#         @Returns
#         ------
#         group : int
#             the number of the groups of the target conv layer.
#         """
#         cpp_conv = list(filter(lambda x: x.kind() == CONV_TYPE, node_group.node_cpps))
#         assert len(cpp_conv) == 1
#         cpp_conv = cpp_conv[0]
#         inputs = list(cpp_conv.inputs())
#         # get the number of the group from the input parameters
#         group = inputs[8].toIValue()
#         return group
#
#     def build(self):
#         """
#         Build the channel dependency for the conv layers
#         in the model. This function return the group number
#         of each conv layers. Note that, here, the group count
#         of conv layers may be larger than their original groups.
#         This is because that the input channel will also be grouped
#         for the group conv layers. To make this clear, assume we
#         have two group conv layers: conv1(group=2), conv2(group=4).
#         conv2 takes the output features of conv1 as input.
#         Then we have to the filters of conv1 can still be
#         divided into 4 groups after filter pruning, because
#         the input channels of conv2 should be divided into
#         4 groups.
#
#         @Returns
#         ------
#         self.dependency : dict
#             key: the name of conv layers, value: the minimum value that the number of
#             filters should be divisible to.
#         """
#         for node in self.graph.nodes_py.nodes_op:
#             if node.op_type in ['Conv2d', 'ConvTranspose2d', CONV_TYPE]:
#                 group = self._get_conv_groups(node)
#
#                 if node.name in self.dependency:
#                     # the conv layer whose group is larger than 1 will require that
#                     # it's number of output channel to be divisible by the number of group.
#                     self.dependency[node.name] = max(self.dependency[node.name], group)
#                 else:
#                     self.dependency[node.name] = group
#                 if group > 1:
#                     # for the conv layer whose group is larger than 1, it will require the number
#                     # of output channels of their parent conv layer to be divisible by group.
#                     parent_convs = self._get_parent_convs(node)
#                     for parent in parent_convs:
#                         if parent in self.dependency:
#                             self.dependency[parent] = max(self.dependency[parent], group)
#                         else:
#                             self.dependency[parent] = group
#         return self.dependency
#
#     @property
#     def dependency_sets(self):
#         return self.dependency
#
#     def export(self, filepath):
#         """
#         export the group dependency to a csv file.
#         Each line describes a convolution layer, the
#         first part of each line is the Pytorch module
#         name of the conv layer. The second part of each
#         line is the group count of the filters in this layer.
#         Note that, the group count may be larger than this
#         layers original group number.
#
#         output example:
#         Conv layer, Groups
#         Conv1, 1
#         Conv2, 2
#         Conv3, 4
#         """
#         header = ['Conv Layer Name', 'Group']
#         with open(filepath, 'w') as csvf:
#             csv_w = csv.writer(csvf, delimiter=',')
#             csv_w.writerow(header)
#             for name in self.dependency:
#                 group = self.dependency[name]
#                 csv_w.writerow([name, group])
