"""This module is for parsing dependencies among layers of model"""
import copy

import torch
import numpy as np
from knnslim.utils import pruner_utils
from knnslim.graph import dependency


class ModelAnalyzer:
    def __init__(self, model, inputs=None, input_shape=None):
        """
        Initialization

        @Parameter
        ------
        model : (torch.nn.Module)
            the deep learning model used in analysis
        inputs : (tuple)
            the input data for model
        input_shape : (tuple)
            the shape of single input  for model

        @Example
        ------
            model : (torch.nn.Module)
            inputs : (torch.randn((1, 3, 64, 64)), torch.randn(1, 1), )
            input_shape : (1, 3, 64, 64)

        """
        self.model = model
        self.input_shape = input_shape
        self.input = inputs
        self.weight_to_module = dict()
        self.layer_to_module = dict()
        self.layer_weight_dict = dict()
        self.dependencies = dict()
        self.get_model_dict(model)

    def get_model_dict(self, model=None):
        """
        Generate dict
        @Parameter
        ------
        model : (torch.nn.Module)
            the deep learning model used in analysis

        @Return:
        ------
            {weight : module}
            {layer : module}
            {layer : weights}

        @Example:
        ------
            {'module.denseblock.conv1.weight' : Conv2d(3, 60, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)}
            {'denseblock.conv1' : Conv2d(3, 60, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)}
            {'denseblock.conv1' : ['module.denseblock.conv1.weight']}
        """
        model = model or self.model
        weights = model.state_dict()
        if isinstance(model, (torch.nn.DataParallel, torch.nn.parallel.DistributedDataParallel)):
            modules = dict(model.module.named_modules())
            model.name = model.module.name = model.module._get_name()
        else:
            modules = dict(model.named_modules())
            model.name = model._get_name()
        weight_to_module = {}
        layer_to_module = {}
        layer_weight_dict = {}
        for name, param in weights.items():
            start = name.find('.') + 1 if name.startswith('module.') else 0
            end = name.rfind('.')
            key = name[start:end]
            if key in modules.keys():
                weight_to_module[name] = modules[key]
                layer_to_module[key] = modules[key]
            if not param.shape: continue  # we do not need 'num_batches_tracked'
            if key in layer_weight_dict.keys():
                layer_weight_dict[key].append(name)
            else:
                layer_weight_dict[key] = [name]

        self.weight_to_module = weight_to_module
        self.layer_to_module = layer_to_module
        self.layer_weight_dict = layer_weight_dict
        return weight_to_module, layer_to_module, layer_weight_dict

    def get_dependencies(self):
        """
        Get dependency of network between layer and layer. We will use 'add', 'cat',
        'weight' in the structural pruning methods.
        @return:
            {'delete' : dict,
             'append' : dict,
             'add' : dict,
             'cat' : dict,
             'weight' : dict,
             'add_prev_set' : set(),
             'cat_prev_set' : set()}
        
        """
        return self.dependencies

    def create_auto_dependencies(self, input_shape=None):
        """
        Search and create all dependencies of network, such as AddDependency which is related to skip connection,
        CatDependency including 'torch.cat' and 'torch.concat' operations, OutputDependency which is consist of 
        passive relation between conv1 passes the input and conv2 accepts the feature map from conv1.
        @return:
            {'delete' : dict,
             'append' : dict,
             'add' : dict,
             'cat' : dict,
             'weight' : dict,
             'add_prev_set' : set(),
             'cat_prev_set' : set()}
        
        """
        self.model.eval()
        input_shape = input_shape or self.input_shape
        dependencies = {'delete': {}, 'append': {}, 'add': {}, 'cat': {}, 'weight': {}, 'add_prev_set': set(),
                        'cat_prev_set': set()}

        if isinstance(self.model, (torch.nn.DataParallel, torch.nn.parallel.DistributedDataParallel)):
            cpu_model = copy.deepcopy(self.model.module)
        else:
            cpu_model = copy.deepcopy(self.model)

        input = self.input if self.input is not None else torch.rand(input_shape)
        if isinstance(input, tuple):
            output = cpu_model(*input)
        else:
            output = cpu_model(input)
        dp = dependency.Dependency(cpu_model, input)

        print('-' * 100)
        ch_dp = dependency.AddDependency(cpu_model, input, dp.get_traced_model())
        print('-' * 100)
        info = ch_dp.dependency_list()
        for prev_names, add_names, next_names in info:
            prev_weights = map_layer_to_weight(prev_names, self.layer_weight_dict)
            next_weights = map_layer_to_weight(next_names, self.layer_weight_dict)

            dependencies['add_prev_set'].update(prev_weights)
            if prev_weights not in dependencies['add'].keys():
                dependencies['add'][prev_weights] = [add_names, next_weights]
            else:
                dependencies['add'][prev_weights][0] += add_names
                dependencies['add'][prev_weights][1] += next_weights

        ch_dp = dependency.CatDependency(cpu_model, input, dp.get_traced_model())
        info = ch_dp.dependency_list()
        for prev_names, cat_names, next_names in info:
            prev_weights = map_layer_to_weight(prev_names, self.layer_weight_dict)
            next_weights = map_layer_to_weight(next_names, self.layer_weight_dict)

            dependencies['cat_prev_set'].update(prev_weights)
            if prev_weights not in dependencies['cat'].keys():
                dependencies['cat'][prev_weights] = [cat_names, next_weights]
            else:
                dependencies['cat'][prev_weights][0] += cat_names
                dependencies['cat'][prev_weights][1] += next_weights

        ch_dp = dependency.OutputDependency(cpu_model, input, dp.get_traced_model())
        info = ch_dp.dependency_dict()

        # print('-' * 50)
        # print('Output dependency dict')
        # for key, val in info.items():
        #    print(key, val)

        # convert layer_names to weight_names
        weight_info = {}
        for layer_name, next_layers in info.items():
            weight_pairs = map_layer_to_weight(layer_name, self.layer_weight_dict, bias=True)
            weight_name, bias_name = weight_pairs[0]  # consider bias in a layer
            next_weights = map_layer_to_all_weights(next_layers, self.layer_weight_dict, self.layer_to_module)
            if bias_name: next_weights = tuple([bias_name]) + next_weights
            weight_info[weight_name] = next_weights

        # recursively merge conv/depthwise_conv dependencies
        def recur_merge_dp_conv(weight_names, weight_dict):
            weight_list = []
            dp_conv_list = []
            for name in weight_names:
                # if name not in weight_dict: continue  # skip non-param layers
                weight_list.append(name)
                op_type = pruner_utils.get_type(self.weight_to_module[name])
                if op_type == 'DepthwiseConv2D' and name in weight_dict:
                    dp_conv_list.append(name)
                    ext_layers, dp_convs = recur_merge_dp_conv(weight_dict[name], weight_dict)
                    weight_list.extend(ext_layers)
                    dp_conv_list.extend(dp_convs)
            return weight_list, dp_conv_list

        dp_conv_list = []
        temp_info = weight_info.copy()
        for key, values in weight_info.items():
            ext_values, dp_convs = recur_merge_dp_conv(values, weight_info)
            temp_info[key] = tuple(ext_values)
            dp_conv_list.extend(dp_convs)

        dp_conv_list = list(set(dp_conv_list))

        for name in dp_conv_list:
            temp_info.pop(name)
        merge_info = temp_info.copy()

        dependencies['weight'] = merge_info
        del cpu_model
        self.dependencies = dependencies
        self.model.train()
        return dependencies


def map_layer_to_weight(layer_names, map_dict, bias=False):
    if not isinstance(layer_names, (list, tuple)):
        layer_names = pruner_utils.tolist(layer_names)
    weight_list = []
    for name in layer_names:
        weight_name = name
        bias_name = None
        if name in map_dict:
            weight_name = map_dict[name][0] if len(map_dict[name]) == 2 else map_dict[name][0]
            #             if bias:
            #                 print(map_dict[name])
            if weight_name.endswith('bias'):
                weight_name = map_dict[name][1]
            assert weight_name.endswith('weight') or weight_name.endswith('weight_orig')
            for i in map_dict[name]:
                if i.endswith('bias'):
                    bias_name = i
                    break
        #             if len(map_dict[name]) == 2 and bias:
        #                 bias_name = map_dict[name][1]
        #                 #print(name, bias_name, map_dict[name])
        #                 assert bias_name.endswith('bias')
        #             elif len(map_dict[name]) == 4 and bias:
        #                 bias_name = map_dict[name][1]
        #                 u_name = map_dict[name][2]
        #                 v_name = map_dict[name][3]
        if bias:
            weight_list.append(tuple([weight_name, bias_name]))
        else:
            weight_list.append(weight_name)
    return tuple(weight_list)


def map_layer_to_all_weights(layer_names, map_dict, layer_dict):
    if not isinstance(layer_names, (list, tuple)):
        layer_names = pruner_utils.tolist(layer_names)
    weight_names = []
    # op_types = ['Conv2D', 'DepthwiseConv2D', 'Linear']
    op_types = [
        'Linear', 'LinearQuant',
        'ConvTranspose1d', 'ConvTranspose2d', 'ConvTranspose3d', 'ConvTranspose2dQuant',
        'Conv1d', 'Conv2D', 'Conv3d', 'DepthwiseConv2D', 'Conv2dQuant', 'DepthwiseConv2DQuant'
    ]
    for name in layer_names:
        if name in map_dict.keys():
            layer_type = pruner_utils.get_type(layer_dict[name])
            # print(layer_type)
            if layer_type in op_types:
                weight_names.append(map_dict[name][0])
            else:
                weight_names.extend(map_dict[name])
    return tuple(weight_names)


def shrink_weights(weight_to_module, state_dict):
    """
        reduce the dimension of each weight according to the zero channels
    @param weight_to_module: dict of {weight_name: module}
    @param state_dict: dict of {weight_name: params}

    @return:
        the modified state_dict
    """
    device = str(list(state_dict.values())[0].device)

    channel_list = []  # [Cout, Cin] contains all Conv, DepthwiseConv, and FC
    prev_ch_zero_indices = []  # only for running mean/variance
    pruned_indices = {}
    for i, (name, weight) in enumerate(state_dict.items()):
        if '_ops' in name or '_params' in name: continue
        value = weight.detach().cpu().numpy()
        layer = weight_to_module[name]
        # Conv weight
        if len(value.shape) == 4:
            # Cin
            vector = np.sum(value, axis=(0, 2, 3))
            zero_indices = np.where(vector == 0)[0]
            value = np.delete(value, zero_indices, axis=1)
            pruned_indices['.'.join(name.split('.')[:-1]) + '_in'] = zero_indices
            # Cout
            vector = np.sum(value, axis=(1, 2, 3))
            zero_indices = np.where(vector == 0)[0]
            value = np.delete(value, zero_indices, axis=0)
            channel_list.append([value.shape[0], value.shape[1]])
            pruned_indices['.'.join(name.split('.')[:-1]) + '_out'] = zero_indices
        # FC weight
        elif len(value.shape) == 2:
            # Cin
            vector = np.sum(value, axis=0)
            zero_indices = np.where(vector == 0)[0]
            value = np.delete(value, zero_indices, axis=1)
            pruned_indices['.'.join(name.split('.')[:-1]) + '_in'] = zero_indices
            # Cout
            vector = np.sum(value, axis=1)
            zero_indices = np.where(vector == 0)[0]
            value = np.delete(value, zero_indices, axis=0)
            channel_list.append([value.shape[0], value.shape[1]])
            pruned_indices['.'.join(name.split('.')[:-1]) + '_out'] = zero_indices
        # bias, Batchnorm (gamma, beta, running_mean/var)
        elif len(value.shape) == 1:
            if 'BatchNorm' in pruner_utils.get_type(layer) and 'weight_u' not in name and 'weight_v' not in name:
                vector = value
                zero_indices = np.where(vector == 0)[0]
                value = np.delete(value, zero_indices, axis=0)
            else:
                continue
                zero_indices = prev_ch_zero_indices
                value = np.delete(value, zero_indices, axis=0)
        else:
            continue
        prev_ch_zero_indices = zero_indices

        # update state_dict in torch model
        if 'cuda' in device:
            final_value = torch.from_numpy(value).cuda()
        elif 'cpu' in device:
            final_value = torch.from_numpy(value)

        state_dict[name] = final_value
        # if 'weight_orig' in name:
        #    layer.weight_orig.data = final_value
        # elif 'weight' in name:
        #    layer.weight.data = final_value
        # elif 'bias' in name:
        #    layer.bias.data = final_value
        # elif 'running_mean' in name:
        #    layer.running_mean.data = final_value
        # elif 'running_var' in name:
        #    layer.running_var.data = final_value

    for i, (name, weight) in enumerate(state_dict.items()):
        value = weight.detach().cpu().numpy()
        layer = weight_to_module[name]
        if len(value.shape) == 1:
            if 'weight_u' in name or 'bias' in name:
                zero_indices = pruned_indices['.'.join(name.split('.')[:-1]) + '_out']
            elif 'weight_v' in name:
                zero_indices = pruned_indices['.'.join(name.split('.')[:-1]) + '_in']

            if 'weight_u' in name or 'weight_v' in name:
                n_rep = np.prod(layer.weight_orig.shape[-2:])
                value = np.delete(value, np.hstack([zero_indices * n_rep + i for i in range(n_rep)]))
            else:
                value = np.delete(value, zero_indices)

            # update state_dict in torch model
            if 'cuda' in device:
                final_value = torch.from_numpy(value).cuda()
            elif 'cpu' in device:
                final_value = torch.from_numpy(value)
            state_dict[name] = final_value

            # if 'weight_u' in name:
            #    layer.weight_u.data = final_value
            # elif 'weight_v' in name:
            #    layer.weight_v.data = final_value
            # elif 'bias' in name:
            #    layer.bias.data = final_value
    return state_dict
