"""This module is for converting a big model to a small model by deleting zero weights and replace related layers"""
import copy
import torch
import torch.nn as nn
import numpy as np

from knnslim.utils import pruner_utils
from knnslim.graph import graph_utils
from knnslim.quant import quant_modules


class SmallModelConverter:
    def __init__(self, model=None, admm_analyzer=None):

        self.model = copy.deepcopy(model)
        self.small_model = copy.deepcopy(model)
        self.admm_analyzer = admm_analyzer

    def converter(self, weight_to_mask=None):
        """
        Find zero weight in the mask and delete them to slim the model.
        
        @Parameters
        ------
        weight_to_mask : (dict)
            the mapping of weight name to pruned mask
        
        @Return
        ------
        small_model : (torch.nn.Module)
            the small model
        """
        mask_keys = weight_to_mask.keys()

        weight_to_module = self.admm_analyzer.weight_to_module
        for wname, module in weight_to_module.items():

            if 'weight' in wname:
                if wname not in mask_keys:
                    mask = torch.ones(module.weight.shape, dtype=torch.float32)
                else:
                    mask = weight_to_mask[wname].cpu().detach()

                # get the type for module
                module_type = pruner_utils.get_type(module)
                print(f'The type of {wname} is {module_type}.')
                # start to slim sub module
                if module_type == 'Conv2D':
                    slim_module = self.slim_conv2d(module, mask)
                    graph_utils.replace_module_by_name(self.small_model, wname.replace('.weight', ''), slim_module)
                elif module_type == 'DepthwiseConv2D':
                    slim_module = self.slim_depthwiseconv2d(module, mask)
                    graph_utils.replace_module_by_name(self.small_model, wname.replace('.weight', ''), slim_module)
                elif module_type == 'BatchNorm2d':
                    slim_module = self.slim_batchnorm2d(module, mask)
                    graph_utils.replace_module_by_name(self.small_model, wname.replace('.weight', ''), slim_module)
                elif module_type == 'Linear':
                    slim_module = self.slim_linear(module, mask)
                    graph_utils.replace_module_by_name(self.small_model, wname.replace('.weight', ''), slim_module)
                elif module_type == 'ConvTranspose2d':
                    slim_module = self.slim_convTranspose2d(module, mask)
                    graph_utils.replace_module_by_name(self.small_model, wname.replace('.weight', ''), slim_module)
                # quant modules
                elif module_type == 'LinearQuant':
                    slim_module = self.slim_linearQuant(module, mask)
                    graph_utils.replace_module_by_name(self.small_model, wname.replace('.weight', ''), slim_module)
                elif module_type == 'Conv2dQuant':
                    slim_module = self.slim_conv2dQuant(module, mask)
                    graph_utils.replace_module_by_name(self.small_model, wname.replace('.weight', ''), slim_module)
                elif module_type == 'DepthwiseConv2DQuant':
                    slim_module = self.slim_depthwiseconv2dQuant(module, mask)
                    graph_utils.replace_module_by_name(self.small_model, wname.replace('.weight', ''), slim_module)
                elif module_type == 'ConvTranspose2dQuant':
                    slim_module = self.slim_convTranspose2dQuant(module, mask)
                    graph_utils.replace_module_by_name(self.small_model, wname.replace('.weight', ''), slim_module)
                    
        return self.small_model

    def slim_conv2d(self, ori_module, mask=None):
        """
        Replace the conv2d module according to mask.
        
        @Parameters
        ------
        ori_module : (torch.nn.Module)
            original conv2d module
        mask : (torch.tensor)
            the mask for conv2d layer
            
        @Return
        ------
        slim_module : (torch.nn.Module)
        """

        # compute the slim out channels
        # cout, cin = mask.shape[0], mask.shape[1]
        mask_1d = torch.mean(mask, (1, 2, 3))
        mask_1d_index = [False if i == 0 else True for i in mask_1d]
        slim_out_channels = np.sum(mask_1d_index, dtype=np.int)

        # compute the slim input channels

        passive_mask_1d = torch.mean(mask, (0, 2, 3))
        passive_mask_1d_index = [False if i == 0 else True for i in passive_mask_1d]
        slim_in_channels = np.sum(passive_mask_1d_index, dtype=np.int)
        # print(mask.shape, slim_out_channels, slim_in_channels)
        if ori_module.bias is None:
            slim_module = nn.Conv2d(slim_in_channels,
                                    slim_out_channels,
                                    kernel_size=ori_module.kernel_size,
                                    stride=ori_module.stride,
                                    padding=ori_module.padding,
                                    groups=ori_module.groups,
                                    bias=False).requires_grad_(True).to(ori_module.weight.device)
        else:
            slim_module = nn.Conv2d(slim_in_channels,
                                    slim_out_channels,
                                    kernel_size=ori_module.kernel_size,
                                    stride=ori_module.stride,
                                    padding=ori_module.padding,
                                    groups=ori_module.groups,
                                    bias=True).requires_grad_(True).to(ori_module.weight.device)
        w_conv = ori_module.weight.clone()
        slim_w_conv = w_conv[mask_1d_index, :, :, :]
        slim_w_conv = slim_w_conv[:, passive_mask_1d_index, :, :]
        slim_module.weight = nn.Parameter(slim_w_conv)

        if ori_module.bias is not None:
            b_conv = ori_module.bias.clone()
            slim_module.bias = nn.Parameter(b_conv[mask_1d_index])

        return slim_module

    def slim_depthwiseconv2d(self, ori_module, mask=None):
        """
        Replace the depthwiseconv2d module according to mask.
        
        @Parameters
        ------
        ori_module : (torch.nn.Module)
            original depthwiseconv2d module
        mask : (torch.tensor)
            the mask for depthwiseconv2d layer
            
        @Return
        ------
        slim_module : (torch.nn.Module)
        """
        # compute the slim output channels
        # passive_cout = mask.shape[0]
        passive_mask_1d = torch.mean(mask, (1, 2, 3))
        passive_mask_1d_index = [False if i == 0 else True for i in passive_mask_1d]
        slim_in_channels = np.sum(passive_mask_1d_index, dtype=np.int)

        # judge bias in ori_module
        if ori_module.bias is None:
            slim_module = nn.Conv2d(slim_in_channels,
                                    slim_in_channels,
                                    kernel_size=ori_module.kernel_size,
                                    stride=ori_module.stride,
                                    padding=ori_module.padding,
                                    groups=slim_in_channels,
                                    bias=False).requires_grad_(True).to(ori_module.weight.device)
        else:
            slim_module = nn.Conv2d(slim_in_channels,
                                    slim_in_channels,
                                    kernel_size=ori_module.kernel_size,
                                    stride=ori_module.stride,
                                    padding=ori_module.padding,
                                    groups=slim_in_channels,
                                    bias=True).requires_grad_(True).to(ori_module.weight.device)

        w_conv = ori_module.weight.clone()
        slim_module.weight = nn.Parameter(w_conv[passive_mask_1d_index, :, :, :])

        if ori_module.bias is not None:
            b_conv = ori_module.bias.clone()
            slim_module.bias = nn.Parameter(b_conv[passive_mask_1d_index])

        return slim_module

    def slim_batchnorm2d(self, ori_module, mask=None):
        """
        Replace the batchnorm2d module according to mask.
        
        @Parameters
        ------
        ori_module : (torch.nn.Module)
            original batchnorm2d module
        mask : (numpy.array)
            the mask for batchnorm2d layer
            
        @Return
        ------
        slim_module : (torch.nn.Module)
        """

        # compute the slim output features
        # passive_cout = mask.shape[0]
        # passive_mask_1d = torch.mean(mask)
        passive_mask_1d_index = [False if i == 0 else True for i in mask]
        slim_num_features = np.sum(passive_mask_1d_index, dtype=np.int)

        slim_module = nn.BatchNorm2d(num_features=slim_num_features,
                                     eps=ori_module.eps,
                                     momentum=ori_module.momentum,
                                     affine=ori_module.affine,
                                     track_running_stats=ori_module.track_running_stats
                                     ).to(ori_module.weight.device)

        w_bn = ori_module.weight.clone()
        running_mean_bn = ori_module.running_mean.clone()
        running_var_bn = ori_module.running_var.clone()
        b_bn = ori_module.bias.clone()

        slim_module.weight = nn.Parameter(w_bn[passive_mask_1d_index])
        slim_module.running_mean = running_mean_bn[passive_mask_1d_index]
        slim_module.running_var = running_var_bn[passive_mask_1d_index]
        slim_module.bias = nn.Parameter(b_bn[passive_mask_1d_index])

        return slim_module

    def slim_linear(self, ori_module, mask=None):
        """
        Replace the linear module according to mask .
        
        @Parameters
        ------
        ori_module : (torch.nn.Module)
            original linear module
        mask : (numpy.array)
            the mask for linear layer
            
        @Return
        ------
        slim_module : (torch.nn.Module)
        """

        # compute the slim out features
        mask_1d = torch.mean(mask, 1)
        mask_1d_index = [False if i == 0 else True for i in mask_1d]
        slim_out_features = np.sum(mask_1d_index, dtype=np.int)

        # compute the slim input features
        passive_mask_1d = torch.mean(mask, 0)
        passive_mask_1d_index = [False if i == 0 else True for i in passive_mask_1d]
        slim_in_features = np.sum(passive_mask_1d_index, dtype=np.int)

        if ori_module.bias is None:
            slim_module = nn.Linear(slim_in_features,
                                    slim_out_features,
                                    bias=False).requires_grad_(True).to(ori_module.weight.device)
        else:
            slim_module = nn.Linear(slim_in_features,
                                    slim_out_features,
                                    bias=True).requires_grad_(True).to(ori_module.weight.device)

        w_linear = ori_module.weight.clone()
        slim_w_linear = w_linear[mask_1d_index, :]
        slim_w_linear = slim_w_linear[:, passive_mask_1d_index]
        slim_module.weight = nn.Parameter(slim_w_linear)
        if slim_module.bias is not None:
            b_linear = ori_module.bias.clone()
            slim_module.bias = nn.Parameter(b_linear[mask_1d_index])
        return slim_module

    def slim_convTranspose2d(self, ori_module, mask=None):
        """
        Replace the convTranspose2d module according to mask .
        
        @Parameters
        ------
        ori_module : (torch.nn.Module)
            original convTranspose2d module
        mask : (numpy.array)
            the mask for convTranspose2d layer
            
        @Return
        ------
        slim_module : (torch.nn.Module)
        """
        # compute the slim out features
        mask_1d = torch.mean(mask, (0, 2, 3))
        mask_1d_index = [False if i == 0 else True for i in mask_1d]
        slim_out_channels = np.sum(mask_1d_index, dtype=np.int)

        # compute the slim input features
        passive_mask_1d = torch.mean(mask, (1, 2, 3))
        passive_mask_1d_index = [False if i == 0 else True for i in passive_mask_1d]
        slim_in_channels = np.sum(passive_mask_1d_index, dtype=np.int)
        # print(slim_out_channels, slim_in_channels)
        if ori_module.bias is None:
            slim_module = nn.ConvTranspose2d(in_channels=slim_in_channels,
                                             out_channels=slim_out_channels,
                                             kernel_size=ori_module.kernel_size,
                                             stride=ori_module.stride,
                                             padding=ori_module.padding,
                                             output_padding=ori_module.output_padding,
                                             groups=ori_module.groups,
                                             bias=False).requires_grad_(True).to(ori_module.weight.device)
        else:
            slim_module = nn.ConvTranspose2d(in_channels=slim_in_channels,
                                             out_channels=slim_out_channels,
                                             kernel_size=ori_module.kernel_size,
                                             stride=ori_module.stride,
                                             padding=ori_module.padding,
                                             output_padding=ori_module.output_padding,
                                             groups=ori_module.groups,
                                             bias=True).requires_grad_(True).to(ori_module.weight.device)
        w_convTran = ori_module.weight
        slim_w_convTran = w_convTran[:, mask_1d_index, :, :]
        slim_w_convTran = slim_w_convTran[passive_mask_1d_index, :, :, :]
        slim_module.weight = nn.Parameter(slim_w_convTran)

        if ori_module.bias is not None:
            b_convTran = ori_module.bias.clone()
            slim_module.bias = nn.Parameter(b_convTran[mask_1d_index])

        return slim_module
    
    def slim_linearQuant(self, ori_module, mask=None):
        """
        Replace the linear quant module according to mask .
        
        @Parameters
        ------
        ori_module : (torch.nn.Module)
            original linear quant module
        mask : (numpy.array)
            the mask for linear layer
            
        @Return
        ------
        slim_module : (torch.nn.Module)
        """

        # compute the slim out features
        mask_1d = torch.mean(mask, 1)
        mask_1d_index = [False if i == 0 else True for i in mask_1d]
        slim_out_features = np.sum(mask_1d_index, dtype=np.int)

        # compute the slim input features
        passive_mask_1d = torch.mean(mask, 0)
        passive_mask_1d_index = [False if i == 0 else True for i in passive_mask_1d]
        slim_in_features = np.sum(passive_mask_1d_index, dtype=np.int)

        if ori_module.bias is None:
            slim_module = quant_modules.LinearQuant(slim_in_features,
                                    slim_out_features,
                                    bias=False).requires_grad_(True).to(ori_module.weight.device)
        else:
            slim_module = quant_modules.LinearQuant(slim_in_features,
                                    slim_out_features,
                                    bias=True).requires_grad_(True).to(ori_module.weight.device)
        
        w_linear = ori_module.weight.clone()
        slim_w_linear = w_linear[mask_1d_index, :]
        slim_w_linear = slim_w_linear[:, passive_mask_1d_index]
        slim_module.weight = nn.Parameter(slim_w_linear)
        slim_w_copy_linear = torch.zeros_like(slim_module.weight)
        slim_module.weight_copy = nn.Parameter(slim_w_copy_linear)
        
        if slim_module.bias is not None:
            b_linear = ori_module.bias.clone()
            slim_module.bias = nn.Parameter(b_linear[mask_1d_index])
        return slim_module
    
    def slim_conv2dQuant(self, ori_module, mask=None):
        """
        Replace the conv2d quant module according to mask.
        
        @Parameters
        ------
        ori_module : (torch.nn.Module)
            original conv2d quant module
        mask : (torch.tensor)
            the mask for conv2d quant layer
            
        @Return
        ------
        slim_module : (torch.nn.Module)
        """

        # compute the slim out channels
        # cout, cin = mask.shape[0], mask.shape[1]
        mask_1d = torch.mean(mask, (1, 2, 3))
        mask_1d_index = [False if i == 0 else True for i in mask_1d]
        slim_out_channels = np.sum(mask_1d_index, dtype=np.int)

        # compute the slim input channels

        passive_mask_1d = torch.mean(mask, (0, 2, 3))
        passive_mask_1d_index = [False if i == 0 else True for i in passive_mask_1d]
        slim_in_channels = np.sum(passive_mask_1d_index, dtype=np.int)
        # print(mask.shape, slim_out_channels, slim_in_channels)
        if ori_module.bias is None:
            slim_module = quant_modules.Conv2dQuant(slim_in_channels,
                                    slim_out_channels,
                                    kernel_size=ori_module.kernel_size,
                                    stride=ori_module.stride,
                                    padding=ori_module.padding,
                                    groups=ori_module.groups,
                                    bias=False).requires_grad_(True).to(ori_module.weight.device)
        else:
            slim_module = quant_modules.Conv2dQuant(slim_in_channels,
                                    slim_out_channels,
                                    kernel_size=ori_module.kernel_size,
                                    stride=ori_module.stride,
                                    padding=ori_module.padding,
                                    groups=ori_module.groups,
                                    bias=True).requires_grad_(True).to(ori_module.weight.device)
        w_conv = ori_module.weight.clone()
        slim_w_conv = w_conv[mask_1d_index, :, :, :]
        slim_w_conv = slim_w_conv[:, passive_mask_1d_index, :, :]
        slim_module.weight = nn.Parameter(slim_w_conv)
        slim_w_copy_conv = torch.zeros_like(slim_w_conv)
        slim_module.weight_copy = nn.Parameter(slim_w_copy_conv)

        if ori_module.bias is not None:
            b_conv = ori_module.bias.clone()
            slim_module.bias = nn.Parameter(b_conv[mask_1d_index])

        return slim_module
    
    def slim_depthwiseconv2dQuant(self, ori_module, mask=None):
        """
        Replace the depthwiseconv2d quant module according to mask.
        
        @Parameters
        ------
        ori_module : (torch.nn.Module)
            original depthwiseconv2d quant module
        mask : (torch.tensor)
            the mask for depthwiseconv2d quant layer
            
        @Return
        ------
        slim_module : (torch.nn.Module)
        """
        # compute the slim output channels
        passive_mask_1d = torch.mean(mask, (1, 2, 3))
        passive_mask_1d_index = [False if i == 0 else True for i in passive_mask_1d]
        slim_in_channels = np.sum(passive_mask_1d_index, dtype=np.int)

        # judge bias in ori_module
        if ori_module.bias is None:
            slim_module = quant_modules.Conv2dQuant(slim_in_channels,
                                    slim_in_channels,
                                    kernel_size=ori_module.kernel_size,
                                    stride=ori_module.stride,
                                    padding=ori_module.padding,
                                    groups=slim_in_channels,
                                    bias=False).requires_grad_(True).to(ori_module.weight.device)
        else:
            slim_module = quant_modules.Conv2dQuant(slim_in_channels,
                                    slim_in_channels,
                                    kernel_size=ori_module.kernel_size,
                                    stride=ori_module.stride,
                                    padding=ori_module.padding,
                                    groups=slim_in_channels,
                                    bias=True).requires_grad_(True).to(ori_module.weight.device)

        w_conv = ori_module.weight.clone()
        slim_module.weight = nn.Parameter(w_conv[passive_mask_1d_index, :, :, :])
        slim_w_copy_conv = torch.zeros_like(slim_module.weight)
        slim_module.weight_copy = nn.Parameter(slim_w_copy_conv)
        
        if ori_module.bias is not None:
            b_conv = ori_module.bias.clone()
            slim_module.bias = nn.Parameter(b_conv[passive_mask_1d_index])

        return slim_module
    
    def slim_convTranspose2dQuant(self, ori_module, mask=None):
        """
        Replace the convTranspose2d quant module according to mask .
        
        @Parameters
        ------
        ori_module : (torch.nn.Module)
            original convTranspose2d quant module
        mask : (numpy.array)
            the mask for convTranspose2d quant layer
            
        @Return
        ------
        slim_module : (torch.nn.Module)
        """
        # compute the slim out features
        mask_1d = torch.mean(mask, (0, 2, 3))
        mask_1d_index = [False if i == 0 else True for i in mask_1d]
        slim_out_channels = np.sum(mask_1d_index, dtype=np.int)

        # compute the slim input features
        passive_mask_1d = torch.mean(mask, (1, 2, 3))
        passive_mask_1d_index = [False if i == 0 else True for i in passive_mask_1d]
        slim_in_channels = np.sum(passive_mask_1d_index, dtype=np.int)

        if ori_module.bias is None:
            slim_module = quant_modules.ConvTranspose2dQuant(in_channels=slim_in_channels,
                                             out_channels=slim_out_channels,
                                             kernel_size=ori_module.kernel_size,
                                             stride=ori_module.stride,
                                             padding=ori_module.padding,
                                             output_padding=ori_module.output_padding,
                                             groups=ori_module.groups,
                                             bias=False).requires_grad_(True).to(ori_module.weight.device)
        else:
            slim_module = quant_modules.ConvTranspose2dQuant(in_channels=slim_in_channels,
                                             out_channels=slim_out_channels,
                                             kernel_size=ori_module.kernel_size,
                                             stride=ori_module.stride,
                                             padding=ori_module.padding,
                                             output_padding=ori_module.output_padding,
                                             groups=ori_module.groups,
                                             bias=True).requires_grad_(True).to(ori_module.weight.device)
        w_convTran = ori_module.weight
        slim_w_convTran = w_convTran[:, mask_1d_index, :, :]
        slim_w_convTran = slim_w_convTran[passive_mask_1d_index, :, :, :]
        slim_module.weight = nn.Parameter(slim_w_convTran)
        slim_w_copy_convTran = torch.zeros_like(slim_w_convTran)
        slim_module.weight_copy = nn.Parameter(slim_w_copy_convTran)
        
        if ori_module.bias is not None:
            b_convTran = ori_module.bias.clone()
            slim_module.bias = nn.Parameter(b_convTran[mask_1d_index])

        return slim_module